#include "fsio.h"

#include <stdlib.h>
#include <iostream>
#include <windows.h>

FILE *file_open_append(char *fpath) {
	return fopenA(fpath);
}

void file_close(FILE *fp) {
	eclose(fp); //in case of failures, displays a message and exits
	//qfclose(fp);
}

//reads the byte contents of a file, returns -1 in case of errors and the number of read bytes otherwise
long file_readbytecontents(char *fpath, char *buffer) {
	FILE * pFile;
	long lSize = 0;
	long result;
	pFile = qfopen(fpath, "rb");
	if (pFile == NULL) {
		msg("error while trying to read bytes of a file\n");
		return -1;
	}
	//get file size:
	qfseek(pFile, 0, SEEK_END);
	lSize = qftell(pFile);
	//rewind(pFile);
	qfclose(pFile);
	pFile = qfopen(fpath, "rb");
	if (pFile == NULL) {
		msg("error while trying to read bytes of a file\n");
		return -1;
	}

	// copy file into out param buffer:
	//result = fread(buffer, 1, lSize, pFile);
	result = qfread(pFile, buffer, lSize);
	if (result != lSize) {
		msg("error while trying to read bytes of a file\n");
		return -1;
	}

	qfclose(pFile);
	return lSize;
}

void file_write_str(FILE *fp, char *str) {
	ewrite(fp, str, strlen(str)); //in case of failures, displays a message and exits
}

int setup_client_socket(char * server, char * server_port, SOCKET* out_sock) {

	//need to link with Ws2_32.lib, Mswsock.lib, and Advapi32.lib
#pragma comment (lib, "Ws2_32.lib")
#pragma comment (lib, "Mswsock.lib")
#pragma comment (lib, "AdvApi32.lib")

#define DEFAULT_BUFLEN 512

	WSADATA wsaData;
	struct addrinfo *result = NULL, *ptr = NULL, hints;
	int iResult;

	//initialize Winsock
	iResult = WSAStartup(MAKEWORD(2, 2), &wsaData);
	if (iResult != 0) {
		msg("WSAStartup failed with error: %d\n", iResult);
		return -1;
	}

	ZeroMemory(&hints, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;

	//resolve the server address and port
	iResult = getaddrinfo(server, server_port, &hints, &result);
	if (iResult != 0) {
		msg("getaddrinfo failed with error: %d\n", iResult);
		WSACleanup();
		return -1;
	}

	//attempt to connect to an address until one succeeds
	for (ptr = result; ptr != NULL; ptr = ptr->ai_next) {

		//create a SOCKET for connecting to server
		*out_sock = socket(ptr->ai_family, ptr->ai_socktype,
			ptr->ai_protocol);
		if (*out_sock == INVALID_SOCKET) {
			msg("socket creation failed with error: %ld\n", WSAGetLastError());
			WSACleanup();
			return -1;
		}

		// Connect to server.
		iResult = connect(*out_sock, ptr->ai_addr, (int)ptr->ai_addrlen);
		if (iResult == SOCKET_ERROR) {
			closesocket(*out_sock);
			*out_sock = INVALID_SOCKET;
			continue;
		}
		break;
	}

	freeaddrinfo(result);

	if (*out_sock == INVALID_SOCKET) {
		msg("socket unable to connect to server\n");
		WSACleanup();
		return -1;
	}

	return 0;
}

int send_buf_via_socket(SOCKET* sock, char* buf) {

	int iResult;
	iResult = qsend(*sock, (void*)buf, (size_t)strlen(buf)); //IDA SDK func,safer than send, see pronet.h 
	if (iResult == SOCKET_ERROR) {
		printf("send char buffer via socket failed with error: %d\n", WSAGetLastError());
		closesocket(*sock);
		WSACleanup();
		return -1;
	}
	//msg("bytes sent via socket: %ld\n", iResult);
	return 0;
}

//test tcp socket io, return errorcode "-1", else returns "0"
int shutdown_socket(SOCKET* sock) {

	int iResult;
	// shutdown the connection since no more data will be sent
	iResult = shutdown(*sock, SD_SEND);
	if (iResult == SOCKET_ERROR) {
		msg("socket shutdown failed with error: %d\n", WSAGetLastError());
		closesocket(*sock);
		WSACleanup();
		return -1;
	}
	return 0;
}


