#include "fsdbio.h"

#include <stdlib.h>
#include <iostream>
#include <windows.h>
#include <vector>

#include "mysql_connection.h"
#include "mysql_driver.h"
#include <cppconn/driver.h>
#include <cppconn/exception.h>
#include <cppconn/resultset.h>
#include <cppconn/statement.h>
#include <cppconn/prepared_statement.h>

const unsigned int mysql_numconnections = 2;
sql::mysql::MySQL_Driver mysql_driver0;
sql::mysql::MySQL_Driver mysql_driver1;
sql::Connection *mysql_conn0;
sql::Connection *mysql_conn1;
sql::Statement *stmt;
sql::Statement *stmt_transaction;
sql::PreparedStatement *pstmt;
sql::ResultSet *res;

//used to commit natfuncs inserts after each 500 inserts into each shard 
struct _natfuncdata {
	std::string funcname;
	double cenx;
	double ceny;
	double cenz;
	double cenw;
	double cendashx;
	double cendashy;
	double cendashz; 
	double cendashw;
	unsigned int natlib_primkey;
	_natfuncdata(std::string _funcname, double _cenx, double _ceny, double _cenz,
		double _cenw, double _cendashx, double _cendashy, double _cendashz, double _cendashw,
		unsigned int _natlib_primkey) : funcname(_funcname), cenx(_cenx), ceny(_ceny), cenz(_cenz), cenw(_cenw),
		cendashx(_cendashx), cendashy(_cendashy), cendashz(_cendashz), cendashw(_cendashw), natlib_primkey(_natlib_primkey)
	{}
};
typedef struct _natfuncdata natfuncdata_t;
std::vector<natfuncdata_t> natfuncbuf_shard0;
std::vector<natfuncdata_t> natfuncbuf_shard1;

FILE *file_open_append(char *fpath) {
	return fopenA(fpath); 
}

void file_close(FILE *fp) {
	eclose(fp); //in case of failures, displays a message and exits
	//qfclose(fp);
}

//reads the byte contents of a file, returns -1 in case of errors and the number of read bytes otherwise
long file_readbytecontents(char *fpath, char *buffer) {
	FILE * pFile;
	long lSize = 0;
	long result;
	pFile = qfopen(fpath, "rb");
	if (pFile == NULL) {
		msg("error while trying to read bytes of a file\n");
		return -1;
	}
	//get file size:
	qfseek(pFile, 0, SEEK_END);
	lSize = qftell(pFile);
	//rewind(pFile);
	qfclose(pFile);
	pFile = qfopen(fpath, "rb");
	if (pFile == NULL) {
		msg("error while trying to read bytes of a file\n");
		return -1;
	}

	// copy file into out param buffer:
	//result = fread(buffer, 1, lSize, pFile);
	result = qfread(pFile,buffer, lSize);
	if (result != lSize) {
		msg("error while trying to read bytes of a file\n");
		return -1;
	}

	qfclose(pFile);
	return lSize;
}

void file_write_str(FILE *fp, char *str) {
	ewrite(fp, str, strlen(str)); //in case of failures, displays a message and exits
}

int setup_client_socket(char * server, char * server_port, SOCKET* out_sock) {

	//need to link with Ws2_32.lib, Mswsock.lib, and Advapi32.lib
	#pragma comment (lib, "Ws2_32.lib")
	#pragma comment (lib, "Mswsock.lib")
	#pragma comment (lib, "AdvApi32.lib")

	#define DEFAULT_BUFLEN 512

	WSADATA wsaData;
	struct addrinfo *result = NULL, *ptr = NULL, hints;
	int iResult;

	//initialize Winsock
	iResult = WSAStartup(MAKEWORD(2, 2), &wsaData);
	if (iResult != 0) {
		msg("WSAStartup failed with error: %d\n", iResult);
		return -1;
	}

	ZeroMemory(&hints, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;

	//resolve the server address and port
	iResult = getaddrinfo(server, server_port, &hints, &result);
	if (iResult != 0) {
		msg("getaddrinfo failed with error: %d\n", iResult);
		WSACleanup();
		return -1;
	}

	//attempt to connect to an address until one succeeds
	for (ptr = result; ptr != NULL; ptr = ptr->ai_next) {

		//create a SOCKET for connecting to server
		*out_sock = socket(ptr->ai_family, ptr->ai_socktype,
			ptr->ai_protocol);
		if (*out_sock == INVALID_SOCKET) {
			msg("socket creation failed with error: %ld\n", WSAGetLastError());
			WSACleanup();
			return -1;
		}

		// Connect to server.
		iResult = connect(*out_sock, ptr->ai_addr, (int)ptr->ai_addrlen);
		if (iResult == SOCKET_ERROR) {
			closesocket(*out_sock);
			*out_sock = INVALID_SOCKET;
			continue;
		}
		break;
	}

	freeaddrinfo(result);

	if (*out_sock == INVALID_SOCKET) {
		msg("socket unable to connect to server\n");
		WSACleanup();
		return -1;
	}

	return 0;
}

int send_buf_via_socket(SOCKET* sock, char* buf) {

	int iResult;
	iResult = qsend(*sock, (void*)buf, (size_t)strlen(buf)); //IDA SDK func,safer than send, see pronet.h 
	if (iResult == SOCKET_ERROR) {
		msg("send char buffer via socket failed with error: %d\n", WSAGetLastError());
		closesocket(*sock);
		WSACleanup();
		return -1;
	}
	//msg("bytes sent via socket: %ld\n", iResult);
	return 0;
}

//test tcp socket io, return errorcode "-1", else returns "0"
int shutdown_socket(SOCKET* sock) {

	int iResult;
	// shutdown the connection since no more data will be sent
	iResult = shutdown(*sock, SD_SEND);
	if (iResult == SOCKET_ERROR) {
		msg("socket shutdown failed with error: %d\n", WSAGetLastError());
		closesocket(*sock);
		WSACleanup();
		return -1;
	}
	return 0;
}

//returns - 1 in case of errors, number of connections otherwise
int mysql_establish_connections() {

	try {
	
		mysql_conn0 = mysql_driver0.connect("tcp://127.0.0.1:3306", "root", "password");
		mysql_conn0->setSchema("centroids");
		mysql_conn0->setAutoCommit(false);
		mysql_conn1 = mysql_driver1.connect("tcp://127.0.0.1:3307", "root", "password");
		mysql_conn1->setSchema("centroids");
		mysql_conn1->setAutoCommit(false);
	
	}
	catch (sql::SQLException &e) {
			msg("error establishing mysql connections: %s\n", e.what());
			msg("MySQL error code: %i\n", e.getErrorCode());
			msg("SQLState: %s\n", e.getSQLState());
			return -1;
	}
	return mysql_numconnections;
}

int mysql_terminate_connections() {
	try {
		//insert the %500 rest of the natfunc data not yet inserted into the db ...

		//... for shard 0 
		//execute prepared insert statements within a programmed non-auto transaction
		stmt_transaction = mysql_conn0->createStatement();
		stmt_transaction->execute("START TRANSACTION");
		//begin of transaction
		for (int i = 0; i < natfuncbuf_shard0.size(); i++) {
			natfuncdata_t act_natfunc_data = natfuncbuf_shard0.at(i);
			pstmt = mysql_conn0->prepareStatement("insert into natfuncs(id,natlib_id,funcname,cenx,ceny,cenz,cenw,cendashx,cendashy,cendashz,cendashw) \
												 												 		     values(default,?,?,?,?,?,?,?,?,?,?)");
			pstmt->setUInt(1, act_natfunc_data.natlib_primkey);
			pstmt->setString(2, act_natfunc_data.funcname);
			pstmt->setDouble(3, act_natfunc_data.cenx);
			pstmt->setDouble(4, act_natfunc_data.ceny);
			pstmt->setDouble(5, act_natfunc_data.cenz);
			pstmt->setDouble(6, act_natfunc_data.cenw);
			pstmt->setDouble(7, act_natfunc_data.cendashx);
			pstmt->setDouble(8, act_natfunc_data.cendashy);
			pstmt->setDouble(9, act_natfunc_data.cendashz);
			pstmt->setDouble(10, act_natfunc_data.cendashw);
			pstmt->executeUpdate();
			delete pstmt;
		}
		//end of transaction
		mysql_conn0->commit();
		//empty the shard's natfunc data buffer
		natfuncbuf_shard0.clear();
		//cleanup
		delete stmt_transaction;

		// ... for shard 1
		//execute prepared insert statements within a programmed non-auto transaction
		stmt_transaction = mysql_conn1->createStatement();
		stmt_transaction->execute("START TRANSACTION");
		//begin of transaction
		for (int i = 0; i < natfuncbuf_shard1.size(); i++) {
			natfuncdata_t act_natfunc_data = natfuncbuf_shard1.at(i);
			pstmt = mysql_conn1->prepareStatement("insert into natfuncs(id,natlib_id,funcname,cenx,ceny,cenz,cenw,cendashx,cendashy,cendashz,cendashw) \
												  												 												 		     values(default,?,?,?,?,?,?,?,?,?,?)");
			pstmt->setUInt(1, act_natfunc_data.natlib_primkey);
			pstmt->setString(2, act_natfunc_data.funcname);
			pstmt->setDouble(3, act_natfunc_data.cenx);
			pstmt->setDouble(4, act_natfunc_data.ceny);
			pstmt->setDouble(5, act_natfunc_data.cenz);
			pstmt->setDouble(6, act_natfunc_data.cenw);
			pstmt->setDouble(7, act_natfunc_data.cendashx);
			pstmt->setDouble(8, act_natfunc_data.cendashy);
			pstmt->setDouble(9, act_natfunc_data.cendashz);
			pstmt->setDouble(10, act_natfunc_data.cendashw);
			pstmt->executeUpdate();
			delete pstmt;
		}
		//end of transaction
		mysql_conn1->commit();
		//empty the shard's natfunc data buffer
		natfuncbuf_shard1.clear();
		//cleanup
		delete stmt_transaction;

		//mysql_conn0->commit();
		//mysql_conn1->commit();
		mysql_conn0->close();
		mysql_conn1->close();
		delete(mysql_conn0);
		delete(mysql_conn1);

	}
	catch (sql::SQLException &e) {
		msg("error terminating mysql connections: %s\n", e.what());
		msg("MySQL error code: %i\n", e.getErrorCode());
		msg("SQLState: %s\n", e.getSQLState());
		return -1;
	}
	return 0;
}

//returns - 1 in case of errors, returns db-auto-generated primary key of new db entry via out param natlib_primkey_generated,
//returns the number of the shard the entry was stored in (based on lib's numfuncs) via out param db_shard
int mysql_store_natlibdata(char *appname,char *libpath, unsigned int numfuncs,std::string *abi, unsigned int *db_shard, unsigned int *natlib_primkey_generated) {
	//msg("my sql store natlibdata: app %s, libpath %s,numfuncs %u, abi %s\n", appname, libpath, numfuncs, abi->c_str());
	try {

		//sharding via number of library functions
		sql::Connection *shard_conn;
		if (numfuncs%2 == 0) {
			shard_conn = mysql_conn0;
			//return shard via out param db_shard
			*db_shard = 0;
		}
		else {
			shard_conn = mysql_conn1;
			//return shard via out param db_shard
			*db_shard = 1;
		}

		//execute a prepared insert statement within a programmed non-auto transaction
		stmt_transaction = shard_conn->createStatement();
		stmt_transaction->execute("START TRANSACTION");
		//begin of transaction
		pstmt = shard_conn->prepareStatement("insert into natlibs(id,appname,libpath,numfuncs,abi,cluster_groupid,clusterid_cen, clusterid_cendash) values(default,?,?,?,?,?,?,?)");
		pstmt->setString(1, appname); //appname should point to zero-terminated utf-16 string, character set of centroid db utf8 
		pstmt->setString(2, libpath);
		pstmt->setUInt(3, numfuncs);
		pstmt->setString(4, *abi);
		pstmt->setInt(5, -1);
		pstmt->setInt(6, -1);
		pstmt->setInt(7, -1);
		pstmt->executeUpdate();
		//end of transaction
		shard_conn->commit();
		//get last insert id on a per-connection basis, this works as long each ida call is processed by only one worker 
		//and distinct connections 
		stmt = shard_conn->createStatement();
		res = stmt->executeQuery("SELECT LAST_INSERT_ID()");
		res->next();
		//return auto-generated primary key via out param
		*natlib_primkey_generated = res->getUInt(1);
		
	}

	catch (sql::SQLException &e) {
		msg("error storing natlib data: %s\n", e.what());
		msg("MySQL error code: %i\n", e.getErrorCode());
		msg("SQLState: %s\n", e.getSQLState());
		return -1;
	}

	//cleanup
	delete pstmt;
	delete res;
	delete stmt_transaction;

	return 0;
}

//returns -1 in case of errors, caller specifies shard via argument, and the primary key of the repective native lib the funtion belongs to (used to write foreign key)
int mysql_store_natfuncdata(std::string *funcname,double cenx, double ceny, double cenz, double cenw,
	double cendashx, double cendashy, double cendashz, double cendashw, unsigned int db_shard, unsigned int natlib_primkey) {
	//msg("my sql store natfuncdata: funcname %s, shard %u,natlib_primkey %u\n", funcname->c_str(), db_shard, natlib_primkey);
	
	//buffer the data for inserts batches of 500 in vectors per shard
	std::string funcname_temp(*funcname);
	natfuncdata_t natfunc_data(funcname_temp, cenx, ceny, cenz, cenw,
		cendashx, cendashy, cendashz, cendashw, natlib_primkey);
	sql::Connection *shard_conn;
	std::vector<natfuncdata_t> *shard_buf;
	if (db_shard == 0) {
		shard_conn = mysql_conn0;
		natfuncbuf_shard0.push_back(natfunc_data);
		shard_buf = &natfuncbuf_shard0;
		if (natfuncbuf_shard0.size() % 500 != 0) return 0;

	}
	else {
		shard_conn = mysql_conn1;
		natfuncbuf_shard1.push_back(natfunc_data);
		shard_buf = &natfuncbuf_shard1;
		if (natfuncbuf_shard1.size() % 500 != 0) return 0;

	}

	//a batch of 500 for a shard was collected =>
	try {
		//execute prepared insert statements within a programmed non-auto transaction
		stmt_transaction = shard_conn->createStatement();
		stmt_transaction->execute("START TRANSACTION");
		//begin of transaction
		for (int i = 0; i < shard_buf->size(); i++) {
			natfuncdata_t act_natfunc_data = shard_buf->at(i);
			pstmt = shard_conn->prepareStatement("insert into natfuncs(id,natlib_id,funcname,cenx,ceny,cenz,cenw,cendashx,cendashy,cendashz,cendashw) \
												 		     values(default,?,?,?,?,?,?,?,?,?,?)");
			pstmt->setUInt(1, act_natfunc_data.natlib_primkey);
			pstmt->setString(2, act_natfunc_data.funcname);
			pstmt->setDouble(3, act_natfunc_data.cenx);
			pstmt->setDouble(4, act_natfunc_data.ceny);
			pstmt->setDouble(5, act_natfunc_data.cenz);
			pstmt->setDouble(6, act_natfunc_data.cenw);
			pstmt->setDouble(7, act_natfunc_data.cendashx);
			pstmt->setDouble(8, act_natfunc_data.cendashy);
			pstmt->setDouble(9, act_natfunc_data.cendashz);
			pstmt->setDouble(10, act_natfunc_data.cendashw);
			pstmt->executeUpdate();
			delete pstmt;
		}
		//end of transaction
		shard_conn->commit();
		
	}
	catch (sql::SQLException &e) {
		msg("error storing natlib data: %s\n", e.what());
		msg("MySQL error code: %i\n", e.getErrorCode());
		msg("SQLState: %s\n", e.getSQLState());
		//empty the shard's natfunc data buffer
		shard_buf->clear();
		return -1;
	}
	//empty the shard's natfunc data buffer
	shard_buf->clear();
	//cleanup
	delete stmt_transaction;
	return 0;
}
