#include "cfgs.h"
#include "fsio.h"
#include "centroid.h"
#include "binaryinfo.h"

//time measurement
#include <tchar.h>
#include <windows.h>

static void traverse_binary_functions();
static void traverse_binary_entrypoints();
static qflow_chart_t* create_flowchart(func_t *func);
static void delete_flowchart(qflow_chart_t *fc);

FILE *fp;

int IDAP_init(void)
{
	return PLUGIN_KEEP;
}
void IDAP_term(void)
{
	// executed when exiting, clean-up jobs here.
	return;
}
// arg -> the plugin might be passed an integer argument from the plugins.cfg file
void IDAP_run(int arg)
{
	char input_file[MAXSTR];

	#ifdef CDEBUG
	msg("centroids plugin started ...\n");
	#endif

	#ifdef ELF32ARM_ONLY
	char arch[8];
	char abi[128];
	if ((is_elf() != 1) || (is_elf_32bit() != 1)) {
		//#ifdef CDEBUG
		msg("input_file is not elf for arm 32 bit, exit\n");
		//#endif
		return;
	}
	if ((get_architecture(arch) == -1) || strcmp(arch, "ARM")) {
		//#ifdef CDEBUG
		msg("input_file is not elf for arm 32 bit, exit\n");
		//#endif
		return;
	}
	if (elf32arm_getarmabi(abi) < 0) {
		//#ifdef CDEBUG
		msg("error while trying to identify ARM ABI, exit\n");
		//#endif
		return;
	}
	//#ifdef CDEBUG
	msg("input file conforms to 32 bit EFL for ARM, ARM abi according to ELF declarations %s\n", abi);
	//#endif
	#endif

	#ifdef RESULTS_TO_FILE
	char results_file[MAXSTR];
	if (qsnprintf(results_file, sizeof(results_file) - 1,"D:\\IDAFiles\\centroidsplugin%i_results.txt", arg) <= 0) {
		msg("qsnprintf failed to write char array for results file, exit\n");
		return;
	}
	fp = file_open_append(results_file);
	char stf[MAXSTR];
	if (get_input_file_path(input_file, MAXSTR) > 0) {
		if (qsnprintf(stf, sizeof(stf) - 1, "started the centroids plugin with input file %s\n", input_file) <= 0) {
			msg("qsnprintf failed to write char array\n");
		}
		else {
			file_write_str(fp, stf);
		}
	}
	else{
		if (qsnprintf(stf, sizeof(stf) - 1, "started the centroids plugin, failed to retrieve binary name\n") <= 0) {
			msg("qsnprintf failed to write char array\n");
		}
		else {
			file_write_str(fp, stf);
		}
	}
	#endif



	#ifdef CDEBUG
	//time measurement
	LARGE_INTEGER frequency;        // ticks per second
	LARGE_INTEGER time_1, time_2;           // ticks
	double elapsedtime_msecs;
	// get ticks per second 
	QueryPerformanceFrequency(&frequency);
	//start timer
	QueryPerformanceCounter(&time_1);
	#endif

	traverse_binary_functions();
	//traverse_binary_entrypoints()

#ifdef CDEBUG
	// stop timer
	QueryPerformanceCounter(&time_2);
	// compute and print the elapsed time in millisec
	elapsedtime_msecs = (time_2.QuadPart - time_1.QuadPart) * 1000.0 / frequency.QuadPart;
	msg("centroids plugin finished, runtime %lf msecs\n", elapsedtime_msecs);
#endif
#ifdef RESULTS_TO_FILE
	if (qsnprintf(stf, sizeof(stf) - 1, "exit the centroids plugin\n") <= 0) {
		msg("qsnprintf failed to write char array\n");
	}
	else {
		file_write_str(fp, stf);
	}
	//file_close(fp);
	eclose(fp);
#endif
	return;
}

// no usage
char IDAP_comment[] = "centroids";
char IDAP_help[] = "centroids";
// name in edit->plugins menu, can be overridden in plugins.cfg 
char IDAP_name[] = "centroids";
char IDAP_hotkey[] = "Alt-2";
// exported object
plugin_t PLUGIN =
{
	IDP_INTERFACE_VERSION, // IDA version plug-in is written for
	0, // flags (see below)
	IDAP_init,
	IDAP_term,
	IDAP_run,
	IDAP_comment, // unused
	IDAP_help,    // unused
	IDAP_name,    // name in edit->plugins menu
	IDAP_hotkey
};

//traverses the functions found by IDA and initiates centroid calculation for each
static void traverse_binary_functions() {
	//msg("start searching for the binarys functions ...\n");

	//following traversal of code segments isn't necessary for correct function count 
	/*unsigned int num_code_segs = 0;
	for (int s = 0; s < get_segm_qty(); s++) {
	segment_t *seg = getnseg(s);
	if ((seg != NULL) && (seg->type == SEG_CODE)) {
	//msg("code segment found ...\n");
	num_code_segs++;
	}
	}*/

	//unsigned int num_funcs = 0;
	func_t *func_p;
	qstring func_name;
	ssize_t func_name_length;
	qflow_chart_t *func_cfg;
	cfg_centroid_t func_centroid;
	cfg_centroid_t func_centroid_dash;
	char strf[MAXSTR];
	int num_funcs;

	for (size_t i = 0; i < get_func_qty(); i++) {
		func_p = getn_func(i);
		if (func_p == NULL) {
			msg("failed loading a function, continue with next function\n");
			continue;
		}
		else {
			//msg("function loaded, start address %a, end address %a\n", f->startEA, f->endEA);
		}
		num_funcs++;
		func_name_length = 0;
		func_name_length = get_func_name2(&func_name, func_p->startEA);
		if (func_name_length <= 0) {
		#ifdef RESULTS_TO_FILE
			if (qsnprintf(strf, sizeof(strf) - 1, "function with non-identifiable name ...\n") <= 0) {
				msg("qsnprintf failed to write char array\n");
			}
			else {
				file_write_str(fp, strf);
			}
		#endif
			msg("failed querying name of the function\n");
		}
		else {
		#ifdef RESULTS_TO_FILE
			if (qsnprintf(strf, sizeof(strf) - 1, "function %s ...\n", func_name.c_str()) <= 0) {
				msg("qsnprintf failed to write char array\n");
			}
			else {
				file_write_str(fp, strf);
			}
		#endif
		}
		#ifdef CDEBUG
		if (!strcmp(func_name.c_str(), "_main")) {
		#endif
			func_cfg = create_flowchart(func_p);
			if (func_cfg == NULL) {
		#ifdef RESULTS_TO_FILE
				if (qsnprintf(strf, sizeof(strf) - 1, "failed creating the cfg of the function") <= 0) {
					msg("qsnprintf failed to write char array\n");
				}
				else {
					file_write_str(fp, strf);
				}
		#endif
				msg("failed creating the cfg of the function, continue with next function\n");
				continue;
			}
			if (calc_centroid(func_cfg, &func_centroid, &func_centroid_dash) == -1) {
		#ifdef RESULTS_TO_FILE
				if (qsnprintf(strf, sizeof(strf) - 1, "failed calculating the centroid value of the function") <= 0) {
					msg("qsnprintf failed to write char array\n");
				}
				else {
					file_write_str(fp, strf);
				}
		#endif
				msg("failed calculating the centroid value of the function, continue with next function\n");
			}
			else {
		#ifdef CDEBUG
				msg("calculated centroid value to <%lf,%lf,%lf,%i>\n", func_centroid.cx, func_centroid.cy,
					func_centroid.cz, func_centroid.w_bar);
				msg("calculated centroid_dash value to <%lf,%lf,%lf,%i>\n", func_centroid_dash.cx, func_centroid_dash.cy,
					func_centroid_dash.cz, func_centroid_dash.w_bar);
		#endif
		#ifdef RESULTS_TO_FILE
				if (qsnprintf(strf, sizeof(strf) - 1, "calculated centroid value to <%lf,%lf,%lf,%i>\n", func_centroid.cx, func_centroid.cy,
					func_centroid.cz, func_centroid.w_bar) <= 0) {
					msg("qsnprintf failed to write char array\n");
				}
				else {
					file_write_str(fp, strf);
				}
				if (qsnprintf(strf, sizeof(strf) - 1, "calculated centroid_dash value to <%lf,%lf,%lf,%i>\n", func_centroid_dash.cx, func_centroid_dash.cy,
					func_centroid_dash.cz, func_centroid_dash.w_bar) <= 0) {
					msg("qsnprintf failed to write char array\n");
				}
				else {
					file_write_str(fp, strf);
				}
		#endif
			}
			delete_flowchart(func_cfg);
		#ifdef CDEBUG
			break;
		}
		#endif
	}
	#ifdef RESULTS_TO_FILE
	if (qsnprintf(strf, sizeof(strf) - 1, "%i functions loaded and analyzed\n", num_funcs) <= 0) {
		msg("qsnprintf failed to write char array\n");
	}
	else {
		file_write_str(fp, strf);
	}
	#endif
	return;
}

//runs through the  entry points found by IDA
static void traverse_binary_entrypoints() {
	unsigned int num_func_entrypoints = 0;
	unsigned int num_nofunc_entrypoints = 0;
	//msg("start searching for the binarys entrypoints\n");
	size_t num_entrypoints = get_entry_qty();
	for (size_t i = 0; i < num_entrypoints; i++)
	{
		asize_t ord = get_entry_ordinal((int)i);
		ea_t ea = get_entry(ord);
		if (ea == BADADDR) {
			msg("failed loading an entry point, continue searching\n");
			continue;
		}
		else {
			//msg("entry point found at address %a\n", ea);
		}
		char entry_name[MAXSTR];
		if (get_entry_name(ord, entry_name, sizeof(entry_name) == -1)) {
			//msg("failed querying name of the entry point\n");
		}
		else {
			//msg("entry points name is %s\n", entry_name);
		}
		//try to determine if entry point is funtion (or e.g. object)
		if (get_func(ea))
		{
			//func_t *pfn = get_func(ea);
			//msg("the entry point is a function\n");
			num_func_entrypoints++;
			continue;
		}

		//msg("the entry point was not identifiable as a function\n");
		num_nofunc_entrypoints++;
	}
	//msg("finished search for entrypoints - %u entrypoints found, %u identified as functions, %u not\n",
	//num_entrypoints, num_func_entrypoints, num_nofunc_entrypoints);
	return;
}

static qflow_chart_t* create_flowchart(func_t *func) {
	//no external blocks, predecessors needed for loop identification
	//qflow_chart_t *fc = new qflow_chart_t("Flow Chart", NULL, func->startEA, func->endEA, /*FC_NOEXT |*/ FC_PREDS);
	qflow_chart_t *fc = new qflow_chart_t("Flow Chart", func, NULL, NULL,/*FC_APPND |*/ /*FC_NOEXT |*/ FC_PREDS);
	return fc;
}

static void delete_flowchart(qflow_chart_t *fc) {
	delete fc;
}
